import { getJsonz, ItemColor } from "@persagy-web/big/lib";
import { SColumnItem } from "@persagy-web/big/lib/items/floor/SColumnItem";
import { SDoorItem } from "@persagy-web/big/lib/items/floor/SDoorItem";
import { SSpaceItem } from "@persagy-web/big/lib/items/floor/SSpaceItem";
import { SVirtualWallItem } from "@persagy-web/big/lib/items/floor/SVirtualWallItem";
import { SWallItem } from "@persagy-web/big/lib/items/floor/SWallItem";
import { SWindowItem } from "@persagy-web/big/lib/items/floor/SWindowItem";
import { SZoneItem } from "@persagy-web/big/lib/items/floor/ZoneItem";
import { Casement } from "@persagy-web/big/lib/types/floor/Casement";
import { Column } from "@persagy-web/big/lib/types/floor/Column";
import { Door } from "@persagy-web/big/lib/types/floor/Door";
import { Space } from "@persagy-web/big/lib/types/floor/Space";
import { VirtualWall } from "@persagy-web/big/lib/types/floor/VirtualWall";
import { Wall } from "@persagy-web/big/lib/types/floor/Wall";
import { Zone } from "@persagy-web/big/lib/types/floor/Zone";
import { FloorData } from "@persagy-web/big/lib/types/FloorData";
import { SColor } from "@persagy-web/draw/lib";
import { SGraphItem, SGraphScene } from "@persagy-web/graph/lib";
import zoneColor from "./config/zoneColor";

/**
 * 楼层平面图
 */
export class FloorScene extends SGraphScene {
    /** json数据  */
    json: string | null = null;
    /** 墙list   */
    wallList: SWallItem[] = [];
    /** 柱子list   */
    columnList: SColumnItem[] = [];
    /** 门list   */
    doorList: SDoorItem[] = [];
    /** 窗list   */
    casementList: SWindowItem[] = [];
    /** 虚拟墙list   */
    virtualWallList: SVirtualWallItem[] = [];
    /** 空间list   */
    spaceList: SSpaceItem[] = [];
    /** 业务空间list */
    zoneList: SZoneItem[] = [];

    /** 空间是否可选  */
    _isSpaceSelectable: boolean = true;
    get isSpaceSelectable(): boolean {
        return this._isSpaceSelectable;
    } // Get isSpaceSelectable
    set isSpaceSelectable(v: boolean) {
        if (this._isSpaceSelectable === v) {
            return;
        }
        this._isSpaceSelectable = v;
        this.spaceList.map(
            (t: SSpaceItem): SSpaceItem => {
                t.selectable = this._isSpaceSelectable;
                return t;
            }
        );
    } // Set isSpaceSelectable

    /** 空间是否可选  */
    _isWallSelectable: boolean = false;
    get isWallSelectable(): boolean {
        return this._isWallSelectable;
    } // Get isWallSelectable
    set isWallSelectable(v: boolean) {
        if (this._isWallSelectable === v) {
            return;
        }
        this._isWallSelectable = v;
        this.wallList.map(
            (t: SWallItem): SWallItem => {
                t.selectable = this._isWallSelectable;
                return t;
            }
        );
    } // Set isWallSelectable

    /** 业务空间是否可选  */
    _isZoneSelectable: boolean = true;
    get isZoneSelectable(): boolean {
        return this._isZoneSelectable;
    } // Get isZoneSelectable
    set isZoneSelectable(v: boolean) {
        if (this._isZoneSelectable === v) {
            return;
        }
        this._isZoneSelectable = v;
        this.zoneList.map(
            (t: SZoneItem, i: number): SZoneItem => {
                t.selectable = this._isZoneSelectable;
                if (v) {
                    t.fillColor = new SColor(zoneColor[i % zoneColor.length]);
                } else {
                    t.fillColor = ItemColor.zoneUnselectColor;
                }
                return t;
            }
        );
    } // Set isZoneSelectable

    constructor() {
        super()
        // 设置始终不出现选择器
        this.selectContainer.showSelect = false;
        // 绑定选择器事件
        this.selectContainer.connect("listChange", this, this.listChange);
    }

    /**
     *  获取底图压缩文件
     *
     *  @param  url     请求数据文件路径
     */
    loadUrl(url: string): Promise<void> {
        return getJsonz(url, true).then((res: any) => {
            this.addBaseMapItem(res.EntityList ? res.EntityList[0].Elements : res.entityList[0].Elements)
            this.json = JSON.stringify(res);
            return res;
        })
    } // Function loadUrl()

    /**
     *  增添所有底图item;
     *
     *  @param  data    itemList对象
     */
    private addBaseMapItem(data: FloorData): void {
        if (data.Walls) {
            data.Walls.forEach((t: Wall): void => {
                this.addWall(t);
            });
        }
        if (data.Columns) {
            data.Columns.forEach((t: Column): void => {
                this.addColumn(t);
            });
        }
        if (data.Windows) {
            data.Windows.forEach((t: Casement): void => {
                this.addCasement(t);
            });
        }
        if (data.VirtualWalls) {
            data.VirtualWalls.forEach((t: VirtualWall): void => {
                this.addVirtualWall(t);
            });
        }
        if (data.Doors) {
            data.Doors.forEach((t: Door): void => {
                this.addDoor(t);
            });
        }
        if (data.Spaces) {
            data.Spaces.forEach((t: Space): void => {
                this.addSpace(t);
            });
        }
    } // Function addBaseMapItem()

    /**
     *  添加空间到scene 中
     *
     *  @param  space   空间item
     */
    addSpace(space: Space): void {
        let item = new SSpaceItem(null, space);
        item.likeCtrl = true;
        item.selectable = this.isSpaceSelectable;
        item.fillColor = ItemColor.spaceColor;
        item.strokeColor = ItemColor.spaceBorderColor;
        this.spaceList.push(item);
        this.addItem(item);
    } // Function addSpace()

    /**
     *  添加柱子到scene 中
     *
     *  @param  column   柱子item
     */
    addColumn(column: Column): void {
        let item = new SColumnItem(null, column);
        item.fillColor = ItemColor.columnColor;
        this.columnList.push(item);
        this.addItem(item);
    } // Function addColumn()

    /**
     *  添加墙到scene 中
     *
     *  @param  wall   墙item
     */
    addWall(wall: Wall): void {
        let item = new SWallItem(null, wall);
        item.fillColor = ItemColor.wallColor;
        // item.selectable = true
        this.wallList.push(item);
        this.addItem(item);
    } // Function addWall()

    /**
     *  添加所有虚拟墙到scene中
     *
     *  @param  virtualWall   虚拟墙item
     */
    addVirtualWall(virtualWall: VirtualWall): void {
        let item = new SVirtualWallItem(null, virtualWall);
        this.virtualWallList.push(item);
        this.addItem(item);
    } // Function addVirtualWall()

    /**
     *  添加门到scene 中
     *
     * @param door
     */
    addDoor(door: Door): void {
        let item = new SDoorItem(null, door);
        this.doorList.push(item);
        this.addItem(item);
    } // Function addDoor()

    /**
     *  添加窗户到scene 中
     *
     * @param casement
     */
    addCasement(casement: Casement): void {
        let item = new SWindowItem(null, casement);
        this.casementList.push(item);
        this.addItem(item);
    } // Function addCasement()

    /**
     *  添加业务空间到scene 中
     *
     *  @param  zone   业务空间list
     */
    addZoneList(zone: Zone[]): void {
        zone.forEach((t, i): void => {
            this.addZone(t, i);
        });
    } // Function addZoneList()

    /**
     *  添加业务空间到scene 中
     *
     *  @param  zone   业务空间
     */
    addZone(zone: Zone, i: number): void {
        let item = new SZoneItem(null, zone);
        item.selectable = this.isZoneSelectable;
        item.fillColor = new SColor(zoneColor[i % zoneColor.length])
        this.zoneList.push(item);
        this.addItem(item);
    } // Function addZone()

    /**
     * 选择器触发事件
    */
    listChange(item: SGraphItem, data: SGraphItem[]): void {
        this.changeSelect(item, data)
    }
    changeSelect(item: SGraphItem, data: SGraphItem[]): void { }

    /**
     * 初始化空间颜色
    */
    initSpaceColor(): void {
        this.spaceList.forEach(
            (t: SSpaceItem) => {
                t.fillColor = ItemColor.spaceColor;
            }
        );
    }

    /**
      * 初始化业务空间颜色
     */
    initZoneColor(): void {
        this.zoneList.forEach(
            (t: SZoneItem, i: number) => {
                if (t.selectable) {
                    t.fillColor = new SColor(zoneColor[i % zoneColor.length]);
                } else {
                    t.fillColor = ItemColor.zoneUnselectColor;
                }
            }
        );
    }

    /**
     * 初始化墙颜色
    */
    initWallColor(): void {
        this.wallList.forEach((t: SWallItem, i: number) => {
            t.fillColor = ItemColor.wallColor;
        })
    }

    /**
     * 清空业务空间
     */
    removeAllZone() {
        this.zoneList.forEach(t => {
            this.removeItem(t);
        })
        this.zoneList = [];
    }
}